/* 
   WMShelf - Workspace Manager Shelf
   Inherits From:	NXIconView
   Declared In:		WMShelf.h
*/

#import <math.h>
#import <AppKit/AppKit.h>
#import <GNUstepGUI/GSDragView.h>

#import <NXFoundation/NXDefaults.h>
#import <NXFoundation/NXFileManager.h>

#import <Preferences/Shelf/ShelfPrefs.h>
#import <Controller.h>
#import <Viewers/PathIcon.h>
#import <Viewers/PathView.h>

#import "Recycler.h"
#import "WMShelf.h"

@implementation WMShelf

- (void)dealloc
{
  [[NSNotificationCenter defaultCenter] removeObserver:self];
  [super dealloc];
}

// --- NXIconView overridings

- initWithFrame:(NSRect)rect
{
  [super initWithFrame:rect];

  autoAdjustsToFitIcons = NO;
  allowsMultipleSelection = NO;
  allowsEmptySelection = YES;
  allowsArrowsSelection = NO;
  allowsAlphanumericSelection = NO;
  
  [self setDelegate:self];
  // [self setTarget:self];
  // [self setAction:@selector(iconClicked:)];
  // [self setDoubleAction:@selector(iconDoubleClicked:)];
  [self setDragAction:@selector(iconDragged:event:)];
  [self registerForDraggedTypes:@[NSFilenamesPboardType]];

  [[NSNotificationCenter defaultCenter]
    addObserver:self
       selector:@selector(iconSlotWidthChanged:)
           name:ShelfIconSlotWidthDidChangeNotification
         object:nil];

  return self;
}

- (void)putIcon:(NXIcon *)anIcon
       intoSlot:(NXIconSlot)aSlot
{
  [super putIcon:anIcon intoSlot:aSlot];
  [anIcon setEditable:NO];
}

- (PathIcon *)iconInSlot:(NXIconSlot)slot
{
  return (PathIcon *)[super iconInSlot:slot];
}

// --- NSView overridings

- (BOOL)acceptsFirstResponder
{
  // Do nothing. It prevents first responder stealing.
  return NO;
}

//=============================================================================
// Shelf additions
//=============================================================================

- (void)reconstructFromRepresentation:(NSDictionary *)aDict
{
  for (NSArray *key in [aDict allKeys]) {
    NSArray    *paths;
    NXIconSlot slot;
    PathIcon   *icon;

    if (![key isKindOfClass:[NSArray class]] || [key count] != 2) {
      continue;
    }
    if (![[key objectAtIndex:0] respondsToSelector:@selector(intValue)] ||
        ![[key objectAtIndex:1] respondsToSelector:@selector(intValue)]) {
      continue;
    }

    slot = NXMakeIconSlot([[key objectAtIndex:0] intValue],
                          [[key objectAtIndex:1] intValue]);
      
    paths = [aDict objectForKey:key];
    if (![paths isKindOfClass:[NSArray class]]) {
      continue;
    }

    icon = [self createIconForPaths:paths];
    if (icon) {
      [icon setDelegate:self];
      [icon registerForDraggedTypes:@[NSFilenamesPboardType]];
      [self putIcon:icon intoSlot:slot];
    }
  }
}

- (NSDictionary *)storableRepresentation
{
  NSMutableDictionary *dict = [[NSMutableDictionary alloc] init];
  Class               nullClass = [NSNull class];

  for (PathIcon *icon in [self icons]) {
    NXIconSlot slot;

    if ([icon isKindOfClass:nullClass] || [[icon paths] count] > 1) {
      continue;
    }

    slot = [self slotForIcon:icon];

    if (slot.x == -1) {
      [NSException raise:NSInternalInconsistencyException
                  format:@"WMShelf: in `-storableRepresentation':"
                   @" Unable to deteremine the slot for an icon that"
                   @" _must_ be present"];
    }

    [dict setObject:[icon paths] forKey:@[[NSNumber numberWithInt:slot.x],
                                          [NSNumber numberWithInt:slot.y]]];
  }

  return [dict autorelease];
}

- (void)checkIfContentsExist
{
  NSFileManager *fm = [NSFileManager defaultManager];
  NSString      *path;

  NSDebugLLog(@"WMShelf", @"[WMShelf] checkShelfContentsExist");

  for (PathIcon *icon in [self icons]) {
    path = [[icon paths] objectAtIndex:0];
    if (![fm fileExistsAtPath:path]) {
      NSLog(@"Shelf element %@ doesn't exist.", path);
      [self removeIcon:icon];
    }
  }
}

- (PathIcon *)createIconForPaths:(NSArray *)paths
{
  PathIcon *icon;
  NSString *path;

  if ((path = [paths objectAtIndex:0]) == nil) {
    return nil;
  }

  icon = [[PathIcon new] autorelease];
  if ([paths count] == 1) {
    [icon setIconImage:[[NSApp delegate] iconForFile:path]];
  }
  [icon setPaths:paths];
  [icon setDoubleClickPassesClick:NO];
  [icon setEditable:NO];
  // [[icon label] setNextKeyView:[[_owner viewer] view]];

  return icon;
}

// --- Notifications
// Generated by WM Preferences
- (void)iconSlotWidthChanged:(NSNotification *)notif
{
  NXDefaults *df = [NXDefaults userDefaults];
  NSSize     size = [self slotSize];
  CGFloat    width = 0.0;

  if ((width = [df floatForKey:ShelfIconSlotWidth]) > 0.0) {
    size.width = width;
  }
  else {
    size.width = SHELF_LABEL_WIDTH;
  }
  [self setSlotSize:size];
}

//=============================================================================
// NXIconView delegate
//=============================================================================
- (void)iconDragged:(id)sender event:(NSEvent *)ev
{
  NSDictionary *iconInfo;
  NSPasteboard *pb = [NSPasteboard pasteboardWithName:NSDragPboard];
  NSRect       iconFrame = [sender frame];
  NSPoint      iconLocation = iconFrame.origin;
  NXIconSlot   iconSlot = [self slotForIcon:sender];

  NSLog(@"[WMShelf] iconDragged: %@", [sender className]);

  draggedSource = self;
  draggedIcon = [sender retain];

  [sender setSelected:NO];

  iconLocation.x += 8;
  iconLocation.y += iconFrame.size.width - 16;

  if (iconSlot.x == 0 && iconSlot.y == 0) {
    draggedMask = NSDragOperationCopy;
  }
  else {
    draggedMask = (NSDragOperationMove|NSDragOperationCopy);
    [self removeIcon:sender];
  }

  // Pasteboard info for 'draggedIcon'
  [pb declareTypes:@[NSFilenamesPboardType] owner:nil];
  [pb setPropertyList:[draggedIcon paths] forType:NSFilenamesPboardType];

  isRootIconDragged = (iconSlot.x == 0 && iconSlot.y == 0) ? YES : NO;
  
  [self dragImage:[draggedIcon iconImage]
               at:iconLocation
           offset:NSZeroSize
            event:ev
       pasteboard:pb
           source:draggedSource
        slideBack:isRootIconDragged];
}

//============================================================================
// Drag and drop
//============================================================================

// --- NSDraggingSource
- (NSDragOperation)draggingSourceOperationMaskForLocal:(BOOL)isLocal
{
  NSLog(@"[WMShelf] draggingSourceOperationMaskForLocal: %@",
        isLocal ? @"YES" : @"NO");
  NXIconSlot iconSlot = [self slotForIcon:[[self selectedIcons] anyObject]];

  if ((iconSlot.x == 0 && iconSlot.y == 0) || isLocal == NO) {
    draggedMask = NSDragOperationCopy;
  }
  else if (isLocal == YES) {
    draggedMask = NSDragOperationMove;
  }
  else {
    draggedMask = NSDragOperationDelete;
  }
  
  return draggedMask;
}

- (void)draggedImage:(NSImage *)image
             beganAt:(NSPoint)screenPoint
{
  dragPoint = screenPoint;
}

- (void)draggedImage:(NSImage *)image
	     endedAt:(NSPoint)screenPoint
	   operation:(NSDragOperation)operation
{
  NSLog(@"[WMShelf] draggedImage:endedAt:operation:%lu mask:%lu",
        operation, draggedMask);

  if ((draggedMask == NSDragOperationCopy) &&
      ![self iconInSlot:lastSlotDragEntered] &&
      isRootIconDragged == NO) {
    NSLog(@"Operation is Copy and no icon in slot [%i,%i]",
          lastSlotDragEntered.x, lastSlotDragEntered.y);
    [self putIcon:draggedIcon intoSlot:lastSlotDragEntered];
    [draggedIcon setDimmed:NO];
    [draggedIcon registerForDraggedTypes:@[NSFilenamesPboardType]];
    [draggedIcon setDelegate:self];    
  }

  [draggedIcon release];
  draggedIcon = nil;
  draggedSource = nil;
  
  lastSlotDragEntered.x = -1;
  lastSlotDragEntered.y = -1;
  lastSlotDragExited.x = -1;
  lastSlotDragExited.y = -1;
  
  draggedMask = NSDragOperationNone;
}

// --- NSDraggingDestination

// - Before the Image is Released

- (BOOL)acceptsDragFromSource:(id)dSource
                    withPaths:(NSArray *)dPaths
{
  if (![dPaths isKindOfClass:[NSArray class]] || [dPaths count] == 0) {
    return NO;
  }
  else if ([dSource isKindOfClass:[Recycler class]]) {
    return NO;
  }

  return YES;
}

- (NSDragOperation)draggingEntered:(id <NSDraggingInfo>)dragInfo
{
  NSPasteboard *pasteBoard = [dragInfo draggingPasteboard];
  NXIconSlot   iconSlot;

  draggedPaths = [pasteBoard propertyListForType:NSFilenamesPboardType];
  draggedSource = [dragInfo draggingSource];
  
  NSLog(@"[WMShelf] -draggingEntered (source:%@)", [draggedSource className]);
  NSLog(@"[WMShelf] -draggingEntered with paths: %@)", draggedPaths);
  
  if ([self acceptsDragFromSource:draggedSource
                        withPaths:draggedPaths] == NO) {
    draggedMask = NSDragOperationNone;
  }
  else {
    ASSIGN(draggedIcon, [self createIconForPaths:draggedPaths]);
    if (draggedIcon == nil) {
      draggedMask = NSDragOperationNone;
    }
    else {
      draggedMask = [self draggingUpdated:dragInfo];
      [draggedIcon setIconImage:[dragInfo draggedImage]];
      [draggedIcon setDimmed:YES];
    }
  }

  return draggedMask;
}

- (NSDragOperation)draggingUpdated:(id <NSDraggingInfo>)dragInfo
{
  NSPoint    mouseLocation;
  NXIconSlot slotUnderMouse;
  NXIcon     *icon = nil;

  // NSLog(@"[WMShelf] -draggingUpdated (source:%@)",
  //       [[dragInfo draggingSource] className]);
  
  mouseLocation = [self convertPoint:[dragInfo draggingLocation] fromView:nil];
  slotUnderMouse = NXMakeIconSlot(floorf(mouseLocation.x / slotSize.width),
                                  floorf(mouseLocation.y / slotSize.height));

  // Draging hasn't leave shelf yet and no slot for drop
  if (slotUnderMouse.x >= slotsWide) {
    slotUnderMouse.x = slotsWide - 1;
  }
  if (slotUnderMouse.y >= slotsTall) {
    slotUnderMouse.y = slotsTall - 1;
  }

  icon = [self iconInSlot:slotUnderMouse];
  if (lastSlotDragEntered.x == slotUnderMouse.x &&
      lastSlotDragEntered.y == slotUnderMouse.y &&
      icon != nil) {
    return draggedMask;
  }
  
  // NSLog(@"DRAG: slot.x,y: %i,%i last slot.x,y: %i,%i slotsWide: %i icon:%@",
  //       slot.x, slot.y, lastSlotDragEntered.x, lastSlotDragEntered.y,
  //       slotsWide, icon);
  
  lastSlotDragEntered.x = slotUnderMouse.x;
  lastSlotDragEntered.y = slotUnderMouse.y;

  draggedMask = NSDragOperationMove;

  if (icon == nil) {
    if ([self acceptsDragFromSource:draggedSource
                          withPaths:draggedPaths] == NO) {
      draggedMask = NSDragOperationNone;
    }
    else if ([draggedSource isKindOfClass:[PathView class]]) {
      draggedMask = NSDragOperationCopy;
    }
    else {
      draggedMask = [dragInfo draggingSourceOperationMask];
    }
    
    if (draggedIcon && [draggedIcon superview]) {
      [self removeIcon:draggedIcon];
    }
    
    if (draggedIcon && draggedMask != NSDragOperationNone) {
      [self putIcon:draggedIcon intoSlot:slotUnderMouse];
    }
  }

  NSLog(@"[WMShelf] draggingUpdated draggedMask=%lu slot: {%i,%i}",
        draggedMask, lastSlotDragEntered.x, lastSlotDragEntered.y);

  return draggedMask;
}

- (void)draggingExited:(id <NSDraggingInfo>)dragInfo
{
  NSLog(@"[WMShelf] -dragginExited (source:%@)", [[dragInfo draggingSource] className]);

  if (draggedIcon && [draggedIcon superview]) {
    [self removeIcon:draggedIcon];
  }
  lastSlotDragExited.x = lastSlotDragEntered.x;
  lastSlotDragExited.y = lastSlotDragEntered.y;
}

// - After the Image is Released

- (BOOL)prepareForDragOperation:(id <NSDraggingInfo>)sender
{
  NSLog(@"[WMShelf] prepare for drag operation = %lu", draggedMask);
  if (draggedMask == NSDragOperationMove) {
    if ([sender draggingSource] == self) {
      NSLog(@"[WMShelf] prepare to Move icon inside Shelf");
    }
    else {
      NSPasteboard *pasteBoard = [sender draggingPasteboard];
      NSArray      *paths = [pasteBoard propertyListForType:NSFilenamesPboardType];
      NSLog(@"[WMShelf] prepare to Move %@ from %@", paths,
            [[sender draggingSource] className]);
    }
  }
  return YES;
}

- (BOOL)performDragOperation:(id <NSDraggingInfo>)sender
{
  NSLog(@"[WMShelf] performDragOperation");
  [draggedIcon registerForDraggedTypes:@[NSFilenamesPboardType]];
  if (delegate) {
    [draggedIcon setDelegate:delegate];
  }
  else {
    [draggedIcon setDelegate:self];
  }

  return YES;
}

- (void)concludeDragOperation:(id <NSDraggingInfo>)sender
{
  NSLog(@"[WMShelf] concludeDragOperation");
  if (draggedIcon && [draggedIcon superview]) {
    [draggedIcon setDimmed:NO];
  }
}

@end
